#----------------------------------------------------------------------------#
# Location:  Plugins/rp_keyframeanimation/transformations.rb
#----------------------------------------------------------------------------#

# Routines for working with transformations
#============================================================================#
module RP::KFA
	
	@@epsilon = 1e-8
	
	# Compute the Euler angle and axis
	def self.euler_angle_axis(t1, t2)  # intial, final
		t1 = Geom::Transformation.new(t1.xaxis, t1.yaxis, t1.zaxis, ORIGIN)
		t2 = Geom::Transformation.new(t2.xaxis, t2.yaxis, t2.zaxis, ORIGIN)
		t = t2 * t1.inverse
		a11, a21, a31 = t.xaxis.to_a
		a12, a22, a32 = t.yaxis.to_a
		a13, a23, a33 = t.zaxis.to_a
		#elems = [a11, a12, a13, a21, a22, a23, a31, a32, a33]
		#puts "%9.3f %9.3f %9.3f \n%9.3f %9.3f %9.3f \n%9.3f %9.3f %9.3f \n" % elems
		#puts
		
		# Compute the angle of rotation (in radians)
		arg = (a11 + a22 + a33 - 1) / 2.0
		if arg >= 1.0
			theta = 0
		elsif arg <= -1.0
			theta = 180.degrees  # i.e. PI radians
		else
			theta = Math.acos arg
		end	
		
		# Find the axis of rotation
		if ( theta < 0.1.degrees )
			theta = 0.0
			e1, e2, e3 = 0, 0, 1  # doesn't matter really
			
		elsif ( theta < 179.9.degrees )
			e1 = (a32 - a23) / (2 * Math.sin(theta))
			e2 = (a13 - a31) / (2 * Math.sin(theta))
			e3 = (a21 - a12) / (2 * Math.sin(theta))
			
		else
			theta = 180.degrees
			if (a11 >= a22) 
				if (a11 >= a33)
					# a11 is maximal diagonal term
					e1 = Math.sqrt(a11 - a22 - a33 + 1) / 2
					halfInverse = 1 / (2 * e1)
					e2 = halfInverse * a12
					e3 = halfInverse * a13
				else 
					# a33 is maximal diagonal term
					e3 = Math.sqrt(a33 - a11 - a22 + 1) / 2
					halfInverse = 1 / (2 * e3)
					e1 = halfInverse * a13
					e2 = halfInverse * a23
				end
			else 
				if (a22 >= a33)
					# a22 is maximal diagonal term
					e2 = Math.sqrt(a22 - a11 - a33 + 1) / 2
					halfInverse = 1 / (2 * e2)
					e1 = halfInverse * a12
					e3 = halfInverse * a23
				else 
					# a33 is maximal diagonal term
					e3 = Math.sqrt(a33 - a11 - a22 + 1) / 2
					halfInverse = 1 / (2 * e3)
					e1 = halfInverse * a13
					e2 = halfInverse * a23
				end
			end		
		end
			
		euler_axis = Geom::Vector3d.new(e1, e2, e3)
		[theta, euler_axis]	
	end
	
	def self.row_reduce(t)
		a11,a21,a31,a41, a12,a22,a32,a42, a13,a23,a33,a43, a14,a24,a34,a44 = t.to_a
		
		a = [
			[a11-1, a12, a13, -a14, 1.0, 0.0, 0.0],
			[a21, a22-1, a23, -a24, 0.0, 1.0, 0.0],
			[a31, a32, a33-1, -a34, 0.0, 0.0, 1.0]
		]
		
		row = 0 
		for col in 0..2
			
			# Find max pivot
			maxrow = row
			for i in (row+1)..2
				maxrow = i  if a[i][col].abs > a[maxrow][col].abs
			end
			a[row], a[maxrow] = a[maxrow], a[row]
			
			# if you get stuck on this column, move to the next
			next  if a[row][col].abs < @@epsilon
			
			# Make pivot value 1
			c = 1 / a[row][col]
			for j in col..6
				a[row][j] *= c
			end
			
			# Eliminate lower column
			for i in (row+1)..2
				c = a[i][col]
				for j in col..6
					a[i][j] -= c * a[row][j]
				end
			end
			
			# Eliminate upper column
			for i in 0..(row-1)
				c = a[i][col]
				for j in col..6
					a[i][j] -= c * a[row][j]
				end
			end
			
			row += 1  
			
		end
		
		# clean up rounding errors
		for i in 0..2
			for j in 0..3
				a[i][j] = 0.0  if a[i][j].abs < 1e-4
			end
		end
		
		return a
	end
	
	
	# Finds the component of translation spanned by the eigenvectors.
	def self.perpendicular_translation(a)
		v = [0.0, 0.0, 0.0]
		
		for i in 0..2
			v[i] = -a[i][3]  if a[i][0].zero? && a[i][1].zero? && a[i][2].zero?
		end
		
		return v  if v[0].zero? && v[1].zero? && v[2].zero?
		
		# Otherwise, we have to find the preimage of v before row reduction
		#puts "computing preimage of v = (#{v[0]}, #{v[1]}, #{v[2]})"
		
		elems = [
			a[0][4], a[1][4], a[2][4], 0.0,
			a[0][5], a[1][5], a[2][5], 0.0,
			a[0][6], a[1][6], a[2][6], 0.0,
			0.0,     0.0,     0.0,     1.0
		]
		
		rr = Geom::Transformation.new elems		
		return rr.inverse * v
	end
	
	
	def self.fixed_point(a)
		x = [0.0, 0.0, 0.0]
		for i in 0..2
			for j in 0..2
				if a[i][j].nonzero?  # i.e. this is the pivot value 1
					x[j] = a[i][3]
					break
				end
			end
		end
		return x
	end
	
	
	# Get scale factor in the x, y, and z direction.
	def self.scale_factors(t1, t2)
		
		# old scale
		a11,a21,a31,a41, a12,a22,a32,a42, a13,a23,a33,a43 = t1.to_a
		sx1 = Math.sqrt(a11*a11 + a21*a21 + a31*a31)
		sy1 = Math.sqrt(a12*a12 + a22*a22 + a32*a32)
		sz1 = Math.sqrt(a13*a13 + a23*a23 + a33*a33)
		
		# new scale
		a11,a21,a31,a41, a12,a22,a32,a42, a13,a23,a33,a43 = t2.to_a
		sx2 = Math.sqrt(a11*a11 + a21*a21 + a31*a31)
		sy2 = Math.sqrt(a12*a12 + a22*a22 + a32*a32)
		sz2 = Math.sqrt(a13*a13 + a23*a23 + a33*a33)
		
		# scale factors
		sx, sy, sz = sx2/sx1, sy2/sy1, sz2/sz1
		
		sx = 1.0  if (sx - 1).abs < @@epsilon
		sy = 1.0  if (sy - 1).abs < @@epsilon
		sz = 1.0  if (sz - 1).abs < @@epsilon
		[sx, sy, sz]
	end

	# The difference between t1 and t2 is decomposed into:
	# a translation,
	# a fixed point,
	# the axis of rotation,
	# the angle of rotation,
	# and the scale factor in the x, y, and z direction.
	def self.diff(t1, t2)
		t = t2 * t1.inverse
		
		# Get angle and axis of rotation
		theta, e = euler_angle_axis(t1, t2)
		#puts "theta = #{ theta.radians } degrees, e = #{ e }"
		
		# Get scale factors
		sx, sy, sz = scale_factors(t1, t2)
		#puts sprintf("scale = %.2f, %.2f, %.2f", sx, sy, sz)
		
		# No Rotation.  Translation and/or Scale only.
		if theta.zero?
			v = t2.origin - t1.origin
			v.length = 0  if v.length < @@epsilon
			#puts "displacement = #{ v }"
			#puts 
			return [v, ORIGIN, e, 0, sx, sy, sz]
		end
		
		# Rotation, possibly with scaling. Hopefully they have the same fixed point.
		a = row_reduce t
		
		p_rot = Geom::Point3d.new(fixed_point(a))
		#puts "fixed point = #{ p_rot }"
		
		v = Geom::Vector3d.new(perpendicular_translation(a))
		v.length = 0  if v.length < @@epsilon
		#puts "translation = (#{ v[0] }, #{ v[1] }, #{ v[2] })"
		#puts
		
		return [v, p_rot, e, theta, sx, sy, sz]
	end
	
end  # module RP::KFA
